#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
import paramiko
import simplejson
import ConfigParser
import logging
import argparse
import socket
import datetime
from launchpadlib.launchpad import Launchpad


def str2bool(str):
    """ simple str to bool converter. """
    return str.lower() in ("yes", "true", "t", "1")


class GerritStream:
    def __init__(self):
        self.port = None
        self.host = None
        self.credfile = None
        self.appname = None
        self.logfile = None
        self.env = None
        self.lp = None
        self.ssh_key = None
        self.cachedir = None
        self.statuses = [
            'In Progress',
            'Fix Committed',
            'New',
            'Confirmed',
            'Fix Released',
        ]
        self.logger = None
        self.logger_formatter = None
        self.tmp_logfile = '/tmp/lpupdatebug.log'
        self.dry_run_enabled = None
        self.bug_regex = r'(?i)(Closes|Partial|Related)-Bug:(?:\s?)(?:#?)(\d*)'
        self.projects = []
        self.do_status_update = False
        self.lp_erro_conn_code = "0x00000lp"
        self.lp_succ_conn_code = "0x11111lp"
        self.ssh_erro_conn_code = "0x00000ssh"
        self.ssh_succ_conn_code = "0x11111ssh"

    def configure(self):
        config_file = '/etc/lpupdatebug/lpupdatebug.conf'
        """ setup logger. """
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger_formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        """ configure console logger when dry-run is enabled. """
        if(self.dry_run_enabled):
            logger_hand_console = logging.StreamHandler()
            logger_hand_console.setFormatter(self.logger_formatter)
            self.logger.addHandler(logger_hand_console)
        try:
            config = ConfigParser.ConfigParser()
            config.read(config_file)
            self.logfile = config.get('lpupdatebug', 'logfile')
            self.port = config.getint('lpupdatebug', 'port')
            self.host = config.get('lpupdatebug', 'host')
            self.credfile = config.get('lpupdatebug', 'credfile')
            self.appname = config.get('lpupdatebug', 'appname')
            self.cachedir = config.get('lpupdatebug', 'cachedir')
            self.ssh_key = config.get('lpupdatebug', 'sshkey')
            self.username = config.get('lpupdatebug', 'username')
            self.env = config.get('lpupdatebug', 'env')
            param_projects = config.get('lpupdatebug', 'projects')
            self.projects = [i.strip() for i in param_projects.split(',')]
            self.do_status_update = str2bool(config.get(
                'lpupdatebug',
                'update_status'
            ))
        except Exception:
            if(self.dry_run_enabled is None):
                if (self.logfile is None):
                    """ something went wrong while parsing cfg file
                        write a logfile to the '/tmp' dir. """
                    self.logfile = self.tmp_logfile
                logger_hand_filetmp = logging.FileHandler(self.logfile)
                logger_hand_filetmp.setFormatter(self.logger_formatter)
                self.logger.addHandler(logger_hand_filetmp)
            self.logger.critical(
                'Critical error checking config file. Exiting. ',
                exc_info=True
            )
            sys.exit(1)
        if(self.dry_run_enabled is None):
            """ config main logger handler. """
            logger_main_handler = logging.FileHandler(self.logfile)
            logger_main_handler.setFormatter(self.logger_formatter)
            self.logger.addHandler(logger_main_handler)

    def lp_auth(self):
        self.lp = Launchpad.login_with(
            self.appname,
            self.env,
            credentials_file=self.credfile,
            launchpadlib_dir=self.cachedir
        )

    def set_status(self, status, bugNumber):
        if(self.do_status_update):
            try:
                bug = self.lp.bugs[bugNumber]
                """ check if there are series in projects for specific bug.
                    if so, we do not update _any_ status. """
                series = True
                try:
                    """ at this point there should not be any exception.
                        if any occurs then smth bad is happening and we might
                        not get proper info about if the bug has series. """
                    for bug_no in bug.bug_tasks:
                        _bug_proj = bug_no.bug_target_name
                        if(_bug_proj not in self.projects):
                            series = True
                            """ found at least one project (series) """
                            break
                        else:
                            series = False
                except:
                    self.logger.error(
                        'Error: series search in set_status failed.',
                        exc_info=True
                    )
                    pass
                """ check if prev we found any series related
                    to our projects. if NOT, then continue. """
                if(not series):
                    """ iterate bug through all related projects
                        if bug is in relation of at least one
                        project, which is present in self.projects
                        then do a LP bug update. """
                    iter = 0
                    for bug_no in bug.bug_tasks:
                        _bug_proj = bug_no.bug_target_name
                        if(_bug_proj in self.projects):
                            task_bug = bug.bug_tasks[iter]
                            if(task_bug.status != status):
                                """ do NOT set to 'In progress' status if
                                    currently bug status is 'Fix committed'
                                    or 'Fix Released'. """
                                if(status == self.statuses[0]
                                   and task_bug.status not in [
                                       self.statuses[1],
                                       self.statuses[4], ]):
                                    if(self.dry_run_enabled):
                                        self.logger.info(
                                            'set_status \n {}'.format(
                                                status,
                                                bugNumber
                                            )
                                        )
                                    else:
                                        task_bug.status = status
                                        task_bug.lp_save()
                                """ set to 'Fix committed' only when
                                    currently bug is set to New, Confirmed
                                    or 'In Progress' status. """
                                if(status == self.statuses[1]
                                   and task_bug.status in [
                                       self.statuses[0],
                                       self.statuses[2],
                                       self.statuses[3], ]):
                                    if(self.dry_run_enabled):
                                        self.logger.info(
                                            'set_status \n {}'.format(
                                                status,
                                                bugNumber
                                            )
                                        )
                                    else:
                                        task_bug.status = status
                                        task_bug.lp_save()
                        iter += 1
            except KeyError:
                """ when here KeyError happens then it means that someone has
                    put wrong LP bugs number or script lacks of permissions """
                pass

    def add_comment(self, comment, subject, bugNumber):
        try:
            bug = self.lp.bugs[bugNumber]
            update = False
            """ iterate bug through all related projects.
                if bug is in relation of at least one
                project, which is present in self.projects
                then do a LP bug update. """
            for bug_no in bug.bug_tasks:
                _bug_proj = bug_no.bug_target_name
                if(_bug_proj in self.projects):
                    update = True
            if(update is True):
                if(self.dry_run_enabled):
                    self.logger.info(
                        'add_comment \n {}'.format(
                            comment,
                            subject,
                            bugNumber
                        )
                    )
                else:
                    bug.newMessage(content=comment, subject=subject)
                    bug.lp_save()
        except KeyError:
            pass

    def change_restored(self, event, bugNumber):
        abandoner = event['restorer']['name']
        abandoner_email = event['restorer']['email']
        branch = event['change']['branch']
        project = event['change']['project']
        review_url = event['change']['url']
        subject = 'Change restored on %s (%s)' % (
            project,
            branch
        )
        comment = 'Change restored by %s <%s> on branch: %s\nReview: %s' % (
            abandoner,
            abandoner_email,
            branch,
            review_url
        )
        try:
            reason = event['reason']
        except KeyError:
            reason = ''
        if(len(reason) > 0):
            comment += '\nReason: %s' % (reason)
        self.add_comment(comment, subject, bugNumber)

    def change_abandoned(self, event, bugNumber):
        abandoner = event['abandoner']['name']
        abandoner_email = event['abandoner']['email']
        branch = event['change']['branch']
        review_url = event['change']['url']
        project = event['change']['project']
        subject = 'Change abandoned on %s (%s)' % (
            project,
            branch
        )
        comment = 'Change abandoned by %s <%s> on branch: %s\nReview: %s' % (
            abandoner,
            abandoner_email,
            branch, review_url
        )
        try:
            reason = event['reason']
        except KeyError:
            reason = ''
        if(len(reason) > 0):
            comment += '\nReason: %s' % (reason)
        self.add_comment(comment, subject, bugNumber)

    def change_merged(self, event, bugNumber, action):
        submitter = event['submitter']['name']
        try:
            submitter_email = event['submitter']['email']
        except KeyError:
            submitter_email = ''
        author = event['patchSet']['author']['name']
        author_email = event['patchSet']['author']['email']
        commit = event['patchSet']['revision']
        commit_message = event['change']['commitMessage']
        date_format = '%c'
        date = datetime.datetime.fromtimestamp(event['patchSet']['createdOn'])
        date = date.strftime(date_format)
        branch = event['change']['branch']
        review_url = event['change']['url']
        project = event['change']['project']
        if(action == 'Related'):
            action = 'Related fix merged'
        else:
            """ set bug to 'Fix Committed' status only when got 'Closes-Bug'.
            """
            if (action == 'Closes'):
                self.set_status(self.statuses[1], bugNumber)
            action = 'Fix merged'
        comment = 'Reviewed: %s\n' % review_url + \
            'Submitter: %s <%s>\n' % (submitter, submitter_email) + \
            'Branch: %s\n\n' % branch + \
            'Commit: %s\n' % commit + \
            'Author: %s <%s>\n' % (author, author_email) + \
            'Date: %s\n\n%s' % (date, commit_message)
        subject = '%s to %s (%s)' % (action, project, branch)
        self.add_comment(comment, subject, bugNumber)

    def patchset_created(self, event, bugNumber, action):
        branch = event['change']['branch']
        author = event['patchSet']['author']['name']
        author_email = event['patchSet']['author']['email']
        review_url = event['change']['url']
        project = event['change']['project']
        if(action == 'Related'):
            action = 'Related fix proposed'
        else:
            action = 'Fix proposed'
        comment = '%s to branch: %s\nChange author: %s <%s>\nReview: %s' % (
            action, branch, author, author_email, review_url
        )
        subject = '%s to %s (%s)' % (action, project, branch)
        self.add_comment(comment, subject, bugNumber)
        """ set bug to 'In Progress' status after adding a new comment. """
        self.set_status(self.statuses[0], bugNumber)

    def main(self):
        """ parse arguments if any. """
        arg_parser = argparse.ArgumentParser()
        arg_parser.add_argument(
            '-D',
            '--dry-run',
            required=False,
            dest='dryrun',
            action='store_true',
            help='Runs in the dry mode.',
        )
        args = arg_parser.parse_args()
        if(args.dryrun):
            self.dry_run_enabled = args.dryrun
        """ parse config file. """
        self.configure()
        """ loop the main part of the script. """
        while True:
            try:
                try:
                    """ auth with LP. """
                    self.lp_auth()
                except socket.error:
                    self.logger.error(
                        'Socket error: Launchpad connection not \
                        established. Reconnecting... | %s' % (
                            self.lp_erro_conn_code
                        ),
                        exc_info=True
                    )
                    continue
                except:
                    self.logger.error(
                        'General error: Unexpected error during LP \
                        connection. Reconnecting... | %s' % (
                            self.lp_erro_conn_code
                        ),
                        exc_info=True
                    )
                    continue
                else:
                    self.logger.info(
                        'LP connection established. | %s' % (
                            self.lp_succ_conn_code
                        ),
                    )
                """ configure ssh connection. """
                client = paramiko.SSHClient()
                pkey = paramiko.RSAKey.from_private_key_file(self.ssh_key)
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                """ try to catch connection refused. """
                try:
                    client.connect(
                        self.host,
                        self.port,
                        username=self.username,
                        pkey=pkey,
                        timeout=30
                    )
                except socket.error:
                    self.logger.error(
                        'Socket error: SSH connection not established. \
                        Reconnecting... | %s' % (
                            self.ssh_erro_conn_code),
                        exc_info=True
                    )
                    continue
                except:
                    self.logger.error(
                        'General error: SSH connection not established. \
                        Reconnecting... | %s' % (
                            self.ssh_erro_conn_code),
                        exc_info=True
                    )
                    continue
                client.get_transport().set_keepalive(60)
                transport = client.get_transport()
                """ are we connected? """
                if(transport.is_active()):
                    self.logger.info('SSH client connected! | %s' % (
                        self.ssh_succ_conn_code)
                    )
                    stdin, stdout, stderr = client.exec_command(
                        'gerrit stream-events'
                    )
                    for line in stdout:
                        try:
                            """ extract from 'change' the commitMessage. """
                            try:
                                event = simplejson.loads(line)
                                change = event['change']
                            except (KeyError, TypeError, NameError):
                                continue
                            """ extract every matchable bugs reference
                                from the commit msg. """
                            bugPattern = re.findall(
                                self.bug_regex,
                                change["commitMessage"]
                            )
                            """ we do bug update on LP only
                                if we have a bug's number. """
                            if len(bugPattern) > 0:
                                """ loop through every bug's reference
                                    for action, bugNumber in bugPattern: """
                                for action, bugNumber in bugPattern:
                                    if bugNumber.isdigit():
                                        if event['type'] == 'patchset-created':
                                            if(event['patchSet']
                                               ['number'] == '1'):
                                                self.patchset_created(
                                                    event,
                                                    bugNumber,
                                                    action
                                                )
                                        if event['type'] == 'change-merged':
                                            self.change_merged(
                                                event,
                                                bugNumber,
                                                action
                                            )
                                        if event['type'] == 'change-abandoned':
                                            self.change_abandoned(
                                                event,
                                                bugNumber,
                                            )
                                        if event['type'] == 'change-restored':
                                            self.change_restored(
                                                event,
                                                bugNumber,
                                            )
                        except Exception:
                            self.logger.error(
                                'Unexpected error in main "line" loop: ',
                                exc_info=True
                            )
                            pass
                    client.close()
                    self.logger.info('SSH client connection closed.')
                else:
                    self.logger.warn('SSH client transport is NOT connected.')
            except Exception:
                self.logger.error(
                    'Unexpected error in main function: ',
                    exc_info=True
                )

gerrit = GerritStream()
gerrit.main()
